/* (c) copyright 2003 David G. Long, Brigham Young University */
/*****************************************************************
  Filename:  slice.c

  This brief programs illustrates how to read a QuikSCAT L1B file
  and generate 3 dB contour slice shapes using BYU-generated
  slice shape files.

  written by DGL at BYU 6/23/2003
  revised by DGL at BYU 2/02/2007 + fixed slice longitude scaling

  note: some sample code to read HDF files is included, though
  the main purpose of the code is to illustrate how to use the shape tables.

  This code should be compiled to include the HDF header directory

******************************************************************/
#include <stdlib.h>  
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <hdf.h>                /* HDF include (local installation) */

#define abs(a) (((a) > 0.0) ? (a) : -(a))

#define slice_start 0           /* slices to look at */
#define slice_stop  8

/* define enum and structure type to store hdf variable arrays efficently */

typedef enum {flt=0, uint4=1, uint2=2, uint1=3} storesize;
typedef struct{
  int rank,dim[3];
  storesize size;
  union {
    float *flt;
    unsigned int *uint4;
    unsigned short int *uint2;
    unsigned char *uint1;
  } data;
} qsdata;


void free_qsdata(qsdata p)  /* free allocated structure memory */
{
  free(p.data.flt);
  p.data.flt=NULL;
}

/* function prototypes for HDF reading routines */

void *get_mem(int32 dt,int32 numval);

void getTextAfterNewline();

qsdata extract_sds(int32 sd_id, char *in_var, int irec, 
		   int slab_size, storesize size);

void read_attrib_byname(int32 sd_id, char *name, int *ntype, 
			int *nval, char *val);

void set_masks(unsigned long *frame_mask, unsigned short *frame_qual_mask, 
	       unsigned long *inst_mask, unsigned short *sigma_mask, 
	       unsigned short *qual_mask, unsigned long *slice_mask);

/* slice shape routine function prototypes */

int load_inst_mode_shape(int inst_mode);

void get_shape(float time, float az, int ib, int islice, float cen_lat,
	       float cen_lon, int *nc, float *cor_lat, float *cor_lon, 
	       float orb_period);


/* global variables */

unsigned short frame_qual_mask;  /* frame_qual_flag */
unsigned short sigma_mask;       /* sigma0_mode_flag */
unsigned short qual_mask;        /* sigma0_qual_flag */
unsigned long  frame_mask;       /* frame_err_status */
unsigned long  inst_mask;        /* frame_inst_status */
unsigned long  slice_mask;       /* slice_qual_flag */

int old_inst_mode = -1;          /* instrument gate width mode, initially flagged as unset */


/* begin main program */

int main(int argc,char *argv[])
{
  char *filename, fn[500], *outpath, op[300], *c, *c2, outfile[500];
  char line[1025], attr_name[512], attr_data2[512], attr_chk[512];
  char values[10240], string[1024];
  int32 retn, attr_index, adata_type, count, sd_id;
  int8 *attr_data;
  int nframes;

  qsdata orbit_time, sigma0_mode_flag, slice_qual_flag;
  qsdata frame_inst_status, frame_err_status, frame_qual_flag, num_pulses;
  qsdata cell_lat, cell_lon, antenna_azimuth, z_vel;
  qsdata slice_lat, slice_lon, slice_sigma0;
  qsdata slice_azimuth, slice_incidence;

  int ir, nframe, rev_num;
  int i, errflag;
  int ntype, nval;

  int iirev, iyears, idays;
  float rev_orbit_period;
  double orb_period;
  
  FILE *fid;

  int iframe, npulses, ii, j, ib, ibb, nc, ascend;

  float cen_lat, cen_lon, ant_az;
  float time_orb, dcen_lat, dcen_lon, dcor_lat, dcor_lon;
  int inst_mode;

  float nlat[8], nlon[8], cx, cy, a[8], b[8], w[32], amin, amax, sigma;
  int isgn, ix, iy, icnt=0;  
  
  
  if (argc < 2) {
    fprintf(stdout,"\nusage: %s L1B_input_name \n",argv[0]);
    fprintf(stdout," input parameters:\n");
    fprintf(stdout,"   L1B_input_name =  input L1B hdf file (full path)\n");
    return(0);
  }

  /* Read the input filename */
  if (argv[1] == NULL){
    printf("Enter name of input L1B file:");
    fgets(line,sizeof(line),stdin);
    sscanf(line,"%s",fn);
    filename=fn;
  } else
    filename=argv[1];
  printf("\nL1B Filename: %s\n",filename);

  /* set global data acceptance flags */
  set_masks(&frame_mask, &frame_qual_mask, &inst_mask, 
	    &sigma_mask, &qual_mask, &slice_mask);

  /* Open the HDF input file and initiate the SD interface */
  sd_id = SDstart(filename, DFACC_RDONLY);
  if (sd_id < 0) {
    printf("*** SDstart error \n");
    exit(-1);
  }

  /* Make sure that the file is a QuikSCAT or SeaWinds Level 2B file */
  attr_index=SDfindattr(sd_id,"ShortName");
  retn=SDattrinfo(sd_id,attr_index,attr_name, &adata_type,&count);
  if (retn < 0) printf("*** error reading ShortName SDattrinfo\n");
  attr_data=(int8 *)malloc(count * DFKNTsize(adata_type));
  retn=SDreadattr(sd_id,attr_index,attr_data);
  if (retn < 0) printf("*** error reading ShortName SDreadattr\n");
  getTextAfterNewline(attr_data,2,attr_data2);
  free(attr_data);
  printf("HDF ShortName: %s\n",attr_data2);
  if (strcmp(attr_data2,"QSCATL1B") != 0 && strcmp(attr_data2,"SWSL1B")  != 0){
    printf("ERROR: Input file is not a QuikSCAT/SeaWinds Level 1B file\n");
    return(0);
  }

  /* read selected vdata header info into strings and decode into variables as needed*/

  read_attrib_byname(sd_id,"rev_number",&ntype,&nval,values);
  getTextAfterNewline(values,2,string);
  iirev=atoi(string);
  rev_num=iirev;
  printf("L1B rev_number: %d\n",iirev);
  
  read_attrib_byname(sd_id,"l1b_actual_frames",&ntype,&nval,values);
  getTextAfterNewline(values,2,string);
  nframes=atoi(string);
  
  read_attrib_byname(sd_id,"RangeBeginningDate",&ntype,&nval,values);
  getTextAfterNewline(values,2,string);
  printf("Begin date: %s\n",string);
  sscanf(string,"%4d-%3d",&iyears,&idays);

  read_attrib_byname(sd_id,"RangeEndingDate",&ntype,&nval,values);
  getTextAfterNewline(values,2,string);
  printf("End date: %s\n",string);

  read_attrib_byname(sd_id,"RangeBeginningTime",&ntype,&nval,values);
  getTextAfterNewline(values,2,string);
  printf("Begin time: %s\n",string);

  read_attrib_byname(sd_id,"RangeEndingTime",&ntype,&nval,values);
  getTextAfterNewline(values,2,string);
  printf("End time: %s\n",string);

  read_attrib_byname(sd_id,"rev_orbit_period",&ntype,&nval,values);
  getTextAfterNewline(values,2,string);
  printf("rev_orbit_period: %s\n",string);
  rev_orbit_period=atof(string);
  sscanf(string,"%lf",&orb_period);
  if (rev_orbit_period < 100.0) rev_orbit_period=6061.0;  
  printf("rev_orbit_period: %s %f\n",string,rev_orbit_period);

  printf("\nL1B file: %s\n",filename);
  printf("Year: %4d  JD: %3d  Rev: %d   L1B frames: %d\n",iyears,idays,iirev,nframes);


  /* Read selected SDSs of the L1B file from frame */

  ir=0;            /* start frame */
  nframe=nframes;  /* end frame to read*/

  printf("\nBegin reading HDF SDSs  Number of frames: %d\n",nframes);
  orbit_time=extract_sds(sd_id,"orbit_time",ir,nframe,uint4);
  frame_inst_status=extract_sds(sd_id,"frame_inst_status",ir,nframe,uint4);
  frame_err_status=extract_sds(sd_id,"frame_err_status",ir,nframe,uint4);
  frame_qual_flag=extract_sds(sd_id,"frame_qual_flag",ir,nframe,uint2);
  num_pulses=extract_sds(sd_id,"num_pulses",ir,nframe,uint1);
  z_vel=extract_sds(sd_id,"z_vel",ir,nframe,flt);
  cell_lat=extract_sds(sd_id,"cell_lat",ir,nframe,flt);
  cell_lon=extract_sds(sd_id,"cell_lon",ir,nframe,flt);
  sigma0_mode_flag=extract_sds(sd_id,"sigma0_mode_flag",ir,nframe,uint2);
  slice_qual_flag=extract_sds(sd_id,"slice_qual_flag",ir,nframe,uint4);
  antenna_azimuth=extract_sds(sd_id,"antenna_azimuth",ir,nframe,flt);
  slice_lat=extract_sds(sd_id,"slice_lat",ir,nframe,flt);
  slice_lon=extract_sds(sd_id,"slice_lon",ir,nframe,flt);
  slice_sigma0=extract_sds(sd_id,"slice_sigma0",ir,nframe,flt);
  slice_azimuth=extract_sds(sd_id,"slice_azimuth",ir,nframe,flt);
  slice_incidence=extract_sds(sd_id,"slice_incidence",ir,nframe,flt);
  printf("End reading SDSs\n");

  /* close access to HDF file */
  SDend(sd_id);  

  /* set old instrument mode value to force change when first valid value available */
  old_inst_mode=-3;
      
  /* open output file */
  fid=fopen("slice.DAT","w");  

  /* for each frame */ 
  for (iframe=ir; iframe < nframe; iframe++) {
    if ((iframe % 100) == 0) printf("Processing frame %d\n",iframe);

    /* only keep valid data */
    if ((frame_err_status.data.uint4[iframe] & frame_mask) == 0 &&
	(frame_inst_status.data.uint4[iframe] & inst_mask) == 0 &&
	(frame_qual_flag.data.uint2[iframe] & frame_qual_mask) == 0) {

      /* get instrument gate width mode at start of frame and see if 
	 appropriate table already loaded */
      inst_mode=((frame_inst_status.data.uint4[iframe] >> 4) % 8);
      if (inst_mode != old_inst_mode) {  /* load new shape table */
	printf(" Changing Instrument Gate Width Mode to %d\n",inst_mode);
	old_inst_mode=inst_mode;
	if (load_inst_mode_shape(inst_mode) < 0) {
	  printf("*** could not load shape table for mode %d\n",inst_mode);
	  exit(-1);
	}
      }

      if (z_vel.data.flt[iframe] > 0.0) 
	ascend=1;
      else
	ascend=0;
      
      /* for each pulse in frame */
      npulses=num_pulses.data.uint1[iframe];
      for (i=0; i<npulses; i++) {
	
	/* keep only valid data */
	if((sigma0_mode_flag.data.uint2[iframe*100+i] & sigma_mask) == 0 &&
	   (slice_qual_flag.data.uint4[iframe*100+i] & slice_mask) == 0) {
	  
	  /* egg center location */
	  cen_lon=cell_lon.data.flt[iframe*100+i];
	  cen_lat=cell_lat.data.flt[iframe*100+i];

	  ant_az = antenna_azimuth.data.flt[iframe*100+i];
	  time_orb=orbit_time.data.uint4[iframe]*0.03125; /* orbit timer in sec */
	  if ((sigma0_mode_flag.data.uint2[iframe*100+i] & 4) == 0)
	    ib=0;   /* inner beam (H pol) */
	  else
	    ib=1;   /* outer beam (V pol) */

	  ibb=ib*2;
	  if (ant_az > 90.0 && ant_az < 270.0) ibb++;  /* aft-looking, otherwise forware-looking */

	  /* for demo, keep only measurements near equator */
	  if (abs(cen_lat) < 1.0) {

	    /* for each slice */
	    for (ii=slice_start; ii<slice_stop; ii++) {

	      /* get sign of sigma-0 measurement */
	      if ((slice_qual_flag.data.uint4[iframe*100+i] & i*4) != 0)
		isgn=-1;
	      else
		isgn=1;
	      /* store value of sigma-0 */
	      sigma=slice_sigma0.data.flt[iframe*800+i*8+ii];

	      /* compute slice center location */
	      cx = fmod(cen_lon+slice_lon.data.flt[iframe*800+i*8+ii]/cos(cen_lat.data * 3.141562954/180.0), 360.0);
	      cy = cen_lat+slice_lat.data.flt[iframe*800+i*8+ii];

	      /* get slice corners */
	      get_shape(time_orb, ant_az, ib, ii, cy, cx, &nc, nlat, nlon, rev_orbit_period);

	      /* print output */
	      fprintf(fid,"%5d %1d %2d %1d %7.2f %1d %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f\n",iframe+1,ascend,i,ii,ant_az,ib,cen_lon,cen_lat,nlon[0],nlat[0],nlon[1],nlat[1],nlon[2],nlat[2],nlon[3],nlat[3],nlon[4],nlat[4],nlon[5],nlat[5],nlon[6],nlat[6],nlon[7],nlat[7]);
	    }
	  }
	}
      }
    }
  }

  fclose(fid);
  
  /* free up storage allocated in the extract_sds routine for each data set */  
  free_qsdata(frame_inst_status);
  free_qsdata(frame_err_status);
  free_qsdata(frame_qual_flag);
  free_qsdata(num_pulses);
  free_qsdata(cell_lat);
  free_qsdata(cell_lon);
  free_qsdata(sigma0_mode_flag);
  free_qsdata(slice_qual_flag);
  free_qsdata(antenna_azimuth);
  free_qsdata(slice_lat);
  free_qsdata(slice_lon);
  free_qsdata(slice_sigma0);
  free_qsdata(slice_azimuth);
  free_qsdata(slice_incidence);

}	

/*****************************************************************
  READ_ATTRIB_BYNAME:  a function to read VDATAs

  1/2003 D.G. Long
*****************************************************************/

void read_attrib_byname(int32 sd_id, char *name,int *ntype,int *nval,char *val)
{
  int32 retn,adata_type,count,attr_index;
  char attr_name[512],tmp[240],*tmp2;
  int rep,rep2;

  /* get index */
  attr_index=SDfindattr(sd_id,name);
  if (attr_index < 0) printf("*** error in SDfindattr read_attrib_byname: %s\n",name);
  
  /* Get information about the file attribute */
  retn=SDattrinfo(sd_id,attr_index,attr_name,&adata_type,&count);
  if (retn < 0) printf("*** error in SDattrinfo read_attrib_byname: %s\n",name);
  *ntype=adata_type;

  /* Read the attribute data */
  retn=SDreadattr(sd_id,attr_index,(int8 *) val);
  if (retn < 0) printf("*** error in SDreadattr read_attrib_byname: %s\n",name);

  /* Obtain number of items in data type array */
  getTextAfterNewline((int8 *) val,1,tmp);
  rep=atoi(tmp);
  tmp2=strchr(tmp,',');
  if (tmp2 != NULL){
    tmp2++;
    rep2=atoi(tmp2);
    rep2=rep*rep2;
  } else
    rep2=rep;
  *nval=rep2;

  /*
  printf("read_attrib_byname: %d %d %d %d %s\n",rep,rep2,count,adata_type,attr_name);
  */

  return;
  
}

/*****************************************************************
  EXTRACT_SDS:  a function to read the data which are stored
	        as Scientific Data Sets (SDS).

  2/1999 K.L. Perry, D.A. Seal
  5/2003 D.G. Long  + made more general - caution: poor conversions
                      are permitted by the code...

*****************************************************************/

qsdata extract_sds(int32 sd_id, char *in_var, int irec, int slab_size, 
		   storesize ss)
{
  char    name[512];
  int32   sds_index,sds_id,retn;
  int32   rank,data_type,dim_sizes[3],num_type,nattrs;
  int32   start[3],edges[3];
  int32   i,numval;
  float64 cal,cal_err,off,off_err;
  void    *out_var;
  int     iflag=0;
  qsdata  caldata;

  int8    *i8;
  uint8   *u8;
  int16   *i16;
  uint16  *u16;
  int32   *i32;
  uint32  *u32;
  float32 *f32;
  float64 *f64;

  dim_sizes[0]=dim_sizes[1]=dim_sizes[2]=0;
  start[0]=irec;
  start[1]=start[2]=0;
  
  /* Search for the index of "in_var" */
  sds_index=SDnametoindex(sd_id,in_var);
  if (sds_index < 0) {
    printf("*** SDS index error %s\n",in_var);
    caldata.data.flt=NULL;
    return(caldata);
  }
  
  /* Select data set corresponding to the returned index */
  sds_id=SDselect(sd_id,sds_index);
  if (sds_id < 1) {
    printf("*** SDS id error %s\n",in_var);
    caldata.data.flt=NULL;
    return(caldata);
  }
  retn=SDgetinfo(sds_id,name,&rank,dim_sizes,&data_type,&nattrs);
  if (retn < 0) {
    printf("*** SDgetinfo error %s %d\n",in_var,retn);
    caldata.data.flt=NULL;
    return(caldata);
  }

  edges[0]=slab_size;
  edges[1]=dim_sizes[1];
  edges[2]=dim_sizes[2];
  
  caldata.rank=rank;
  caldata.dim[0]=dim_sizes[0];
  caldata.dim[1]=dim_sizes[1];
  caldata.dim[2]=dim_sizes[2];

  for (i=0,numval=1;i<rank;i++) numval *=caldata.dim[i];

  /* Allocate the memory for the output variable */
  out_var=get_mem(data_type,numval);
  if (out_var == NULL) {
    printf("*** Memory allocation error %s %d\n",in_var,retn);
    caldata.data.flt=NULL;
    return(caldata);
  }

  /* Read the data set into the "out_var" array */
  retn=SDreaddata(sds_id,start,NULL,edges,(VOIDP)out_var);
  if (retn < 0) {
    printf("*** SDreaddata error %s %d\n",in_var,retn);
    caldata.data.flt=NULL;
    return(caldata);
  }

  /* Get the calibration and offset values of input */
  retn=SDgetcal(sds_id,&cal,&cal_err,&off,&off_err,&num_type);
  if (retn < 0) {
    printf("*** SDgetcal error %s %d\n",in_var,retn);
    caldata.data.flt=NULL;
    return(caldata);
  }

  /* Print Local Attributes */
  /*
    printf("%-19s %-3d %-4d %-4d %-4d %-2.9f %-2.9f %3d\n",
            name,rank,dim_sizes[0],dim_sizes[1],dim_sizes[2], cal,
            off,data_type);
   */

  retn=SDendaccess(sds_id);

  if (cal == 1.0)
    iflag=1; 
  else if (ss != flt)
    printf("** SD calibration error in store size %d %s\n",ss,in_var);
    
  /*
  printf("in_var: %s  %d cal=%f %d  %d %d\n",in_var,data_type,cal,iflag,rank,numval);
  */

  /* Allocate memory for callibrated output */
  switch (ss) {
  case flt:
    caldata.data.flt=malloc(numval*sizeof(float));
    break;
  case uint4:
    caldata.data.flt=malloc(numval*sizeof(unsigned int));
    break;
  case uint2:
    caldata.data.flt=malloc(numval*sizeof(unsigned short));
    break;
  case uint1:
    caldata.data.flt=malloc(numval*sizeof(unsigned char));
  }
  caldata.size = ss;

  if (data_type == 20) {  /* i8 */
    if (iflag==1)
      for (i=0,i8=out_var; i<numval; i++,i8++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *i8;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *i8;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *i8;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *i8;
	}
    else
      for (i=0,i8=out_var; i<numval;i++,i8++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *i8 * cal;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *i8 * cal;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *i8 * cal;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *i8 * cal;
	}
  }
  else if (data_type == 21) { /* u8 */
    if (iflag==1)
      for (i=0,u8=out_var; i<numval; i++,u8++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *u8;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *u8;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *u8;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *u8;
	}
    else
      for (i=0,u8=out_var; i<numval;i++,u8++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *u8 * cal;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *u8 * cal;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *u8 * cal;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *u8 * cal;
	}
  }
  else if (data_type == 22) { /* i16 */
    if (iflag==1)
      for (i=0,i16=out_var; i<numval; i++,i16++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *i16;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *i16;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *i16;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *i16;
	}
    else
      for (i=0,i16=out_var; i<numval;i++,i16++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *i16 * cal;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *i16 * cal;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *i16 * cal;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *i16 * cal;
	}
  }
  else if (data_type == 23) { /* u16 */
    if (iflag==1)
      for (i=0,u16=out_var; i<numval; i++,u16++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *u16;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *u16;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *u16;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *u16;
	}
    else
      for (i=0,u16=out_var; i<numval;i++,u16++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *u16 * cal;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *u16 * cal;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *u16 * cal;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *u16 * cal;
	}
  }
  else if (data_type == 24) { /* i32 */
    if (iflag==1)
      for (i=0,i32=out_var; i<numval; i++,i32++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *i32;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *i32;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *i32;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *i32;
	}
    else
      for (i=0,i32=out_var; i<numval;i++,i32++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *i32 * cal;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *i32 * cal;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *i32 * cal;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *i32 * cal;
	}
  }
  else if (data_type == 25) { /* u32 */ /* potential conversion problem, but OK */
    if (iflag==1)
      for (i=0,u32=out_var; i<numval; i++,u32++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *u32;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *u32;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *u32;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *u32;
	}
    else
      for (i=0,u32=out_var; i<numval;i++,u32++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *u32 * cal;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *u32 * cal;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *u32 * cal;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *u32 * cal;
	}
  }
  else if (data_type == 5) {  /* f32 */
    if (iflag==1)
      for (i=0,f32=out_var; i<numval; i++,f32++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *f32;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *f32;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *f32;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *f32;
	}
    else
      for (i=0,f32=out_var; i<numval;i++,f32++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *f32 * cal;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *f32 * cal;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *f32 * cal;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *f32 * cal;
	}
  }
  else if (data_type == 6) { /* f64 *//* poor conversion, but not used */
    if (iflag==1)
      for (i=0,f64=out_var; i<numval; i++,f64++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *f64;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *f64;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *f64;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *f64;
	}
    else
      for (i=0,f64=out_var; i<numval;i++,f64++)
	switch (ss) {
	case flt:
	  caldata.data.flt[i] = *f64 * cal;
	  break;
	case uint4:
	  caldata.data.uint4[i] = *f64 * cal;
	  break;
	case uint2:
	  caldata.data.uint2[i] = *f64 * cal;
	  break;
	case uint1:
	  caldata.data.uint1[i] = *f64 * cal;
	}
  }
  else{
    printf("Unknown Data Type \n");
  }
  free(out_var);

  return(caldata);
}


/*****************************************************************
   GET_MEM:  a function which allocates the memory for an SDS
              using the data type.

   12/1998 K.L. Perry
           -adapted from get_type.c by A.V. Tran
*****************************************************************/

void *get_mem(int32 dt, int32 numval)
{
  switch(dt) {

  case DFNT_INT8   : return(malloc(numval*sizeof(int8)));
  case DFNT_UINT8  : return(malloc(numval*sizeof(uint8)));
  case DFNT_INT16  : return(malloc(numval*sizeof(int16)));
  case DFNT_UINT16 : return(malloc(numval*sizeof(uint16)));
  case DFNT_INT32  : return(malloc(numval*sizeof(int32)));
  case DFNT_UINT32 : return(malloc(numval*sizeof(uint32)));
  case DFNT_FLOAT32: return(malloc(numval*sizeof(float32)));
  case DFNT_FLOAT64: return(malloc(numval*sizeof(float64)));

  default : return("unknown number type");
  }
}

/*****************************************************************
   getTextAfterNewline: a function which extracts all characters 
                        between two new lines from a string.

   7/1998 D.A. Seal
*****************************************************************/

void getTextAfterNewline( char *inputtext, int nn, char outputtext[240] )
{
  int nni, ii;
  char *ch;
  
  for ( nni = 0, ii = 0, ch = inputtext; *ch != '\0'; ch++ ) {
    if ( *ch == '\n' ) nni++;
    if ( nni == nn && *ch != '\n' ) {
      outputtext[ii] = *ch;
      ii++;
    }
    if ( nni > nn ) break;
  }
  outputtext[ii] = '\0';
  return;
}


/* utility routines */

void set_masks(unsigned long *frame_mask, unsigned short *frame_qual_mask, 
	       unsigned long *inst_mask, unsigned short *sigma_mask, 
	       unsigned short *qual_mask, unsigned long *slice_mask)
{
  int i;
  
  /* set to ones all bits which we will reject if set */

  /* frame_err_status */
  *frame_mask = ~(1<<23 | 1<<24);

  /* frame_qual_flag */
  *frame_qual_mask = ~(1 | 1<<1 | 1<<2 | 1<<3);

  /* frame_inst_status */
  *inst_mask = ~(1<<2 | 1<<4 | 1<<5 | 1<<6 | 1<<7 | 1<<8 | 1<<9 |
		 1<<10 | 1<<11 | 1<<23);  

  /* sigma0_mode_flag */
  *sigma_mask = ~(1<<2 | 1<<3 | 1<<6 | 1<<7 | 1<<8 | 1<<10);  

  /* sigma0_qual_flag */
  *qual_mask = ~(1<<1 | 1<<2 | 1<<3);  

  /* slice_qual_flag */
  for (i=0; i<8; i++)
    *slice_mask = *slice_mask | (1<<(i*4) | 1<<(i*4+1) | 1<<(i*4+2)); 
  *slice_mask = ~*slice_mask;
 
  printf("Masks %X %X %X %X %X %X\n",*frame_mask, *frame_qual_mask, 
	 *inst_mask, *sigma_mask, *qual_mask, *slice_mask);

  /*  
c frame_err_status
c	frame_mask=ibclr(frame_mask,12)    ! clear SES Data Overrun Flag #?#
c	frame_mask=ibclr(frame_mask,13)    ! clear SES Data Underrun Flag #?#
c	frame_mask=ibclr(frame_mask,14)    ! clear Payload Bus Interface (PBI) Flag #?#
	frame_mask=ibclr(frame_mask,23)    ! clear Attitude Data Flag #?#
	frame_mask=ibclr(frame_mask,24)    ! clear Ephemeris Data Flag #?#

c frame_qual_flag
	frame_qual_mask=ibclr(frame_qual_mask,0)    ! clear frame filler bit 1
	frame_qual_mask=ibclr(frame_qual_mask,1)    ! clear frame filler bit 2
	frame_qual_mask=ibclr(frame_qual_mask,2)    ! clear frame CRC flag bit 1
	frame_qual_mask=ibclr(frame_qual_mask,3)    ! clear frame CRC flag bit 2

c frame_inst_status
	inst_mask=ibclr(inst_mask,2)    ! clear first pulse count bit
	inst_mask=ibclr(inst_mask,4)    ! clear gate width bits
	inst_mask=ibclr(inst_mask,5)    ! clear gate width bits
	inst_mask=ibclr(inst_mask,6)    ! clear gate width bits
	inst_mask=ibclr(inst_mask,7)    ! clear hires/egg acquisition flag
	inst_mask=ibclr(inst_mask,8)    ! clear begin calibration bit
	inst_mask=ibclr(inst_mask,9)    ! clear SES select bit
	inst_mask=ibclr(inst_mask,10)   ! clear SAS select bit
	inst_mask=ibclr(inst_mask,11)   ! clear TWTA select bit
	inst_mask=ibclr(inst_mask,23)   ! clear TWTA monitor enable bit #?#

c sigma0_mode_flag
	sigma_mask=ibclr(sigma_mask,2)  ! clear beam flag bit
	sigma_mask=ibclr(sigma_mask,3)  ! clear location bit
	sigma_mask=ibclr(sigma_mask,6)  ! clear gate width bits
	sigma_mask=ibclr(sigma_mask,7)  ! clear gate width bits
	sigma_mask=ibclr(sigma_mask,8)  ! clear gate width bits
	sigma_mask=ibclr(sigma_mask,10) ! clear SES select bit
c	sigma_mask=ibclr(sigma_mask,11) ! clear spin rate bit

c sigma0_qual_flag
c	qual_mask=ibclr(qual_mask,0)  ! clear sigma0 useable bit #?#
	qual_mask=ibclr(qual_mask,1)  ! clear low snr bit #?#
	qual_mask=ibclr(qual_mask,2)  ! clear negative sigma-0 bit
	qual_mask=ibclr(qual_mask,3)  ! clear sigma0 out of range bit #?#

c slice_qual_flag 
	do i=1,8
	   j=(i-1)*4
	   slice_mask=ibclr(slice_mask,j+1) ! clear sign flag bit
	   slice_mask=ibclr(slice_mask,j+0) ! clear gain threshold bit #?#
	   slice_mask=ibclr(slice_mask,j+2) ! clear low SNR flag bit #?#
c	   slice_mask=ibclr(slice_mask,j+3) ! clear location bit #?#
	end do
*/
}




/*********************************************************************************/


/* slice shape code */


  
float rinterp(float xx, float yy, float r1, float r2, float r3, float r4)
{
  return((r2-r1)*yy+r1+((r4-r3)*yy+r3-(r2-r1)*yy-r1)*xx);
}



/* shape table storage area */

static struct {
  int norbs,nazs;
  float orbtimes[36],az_angs[8*36*36*2],corners[16*8*36*36*2];
} shape;


/* load slice shape table into memory */

int load_inst_mode_shape(int mode)
{
  char *loc_path, loc_file[512], loc_file_string[]=".";
  FILE *fid;
  int i,ib,iaz,iorb,islice;

  /* get path to shape table locations from environmental variable BYU_SHAPES_DIR 
     set to to path to local location prior to call */
  if ((loc_path =(char *) getenv("BYU_SHAPES_DIR"))==NULL) {
    loc_path = loc_file_string;
    printf("*** The environment variable BYU_SHAPES_DIR is not set/\n");
  }
  
  sprintf(loc_file,"%s/shape_in%d.tab",loc_path,mode+1);
  printf(" Reading shape table file %s\n",loc_file);
  fid=fopen(loc_file,"r");
  if (fid==NULL) {
    printf("*** could not open shape table file %s\n",loc_file);
    return(-1);
  }
  ib=0;

  fscanf(fid,"%d %d",&shape.norbs,&shape.nazs);

  for (i=0;i<shape.norbs;i++)
    fscanf(fid,"%f",&shape.orbtimes[i]);

  for (iorb=0; iorb<shape.norbs; iorb++)
    for (iaz=0; iaz<shape.nazs; iaz++)
      for (i=0; i<8; i++)
	fscanf(fid,"%f",&shape.az_angs[i+iaz*8+iorb*8*shape.nazs+ib*8*shape.nazs*shape.norbs]);

  for (iorb=0; iorb<shape.norbs; iorb++)
    for (iaz=0; iaz<shape.nazs; iaz++)
      for (islice=0; islice<8; islice++)
	for (i=0; i<16; i++)
	  fscanf(fid,"%f",&shape.corners[i+islice*16+iaz*8*16+iorb*8*16*shape.nazs+ib*8*16*shape.nazs*shape.norbs]);

  fclose(fid);

  sprintf(loc_file,"%s/shape_out%d.tab",loc_path,mode+1);
  printf(" Reading shape table file %s\n",loc_file);
  fid=fopen(loc_file,"r");
  if (fid==NULL) {
    printf("*** could not open shape table file %s\n",loc_file);
    return(-1);
  }
  ib=1;

  fscanf(fid,"%d %d",&shape.norbs,&shape.nazs);

  for (i=0;i<shape.norbs;i++)
    fscanf(fid,"%f",&shape.orbtimes[i]);

  for (iorb=0; iorb<shape.norbs; iorb++)
    for (iaz=0; iaz<shape.nazs; iaz++)
      for (i=0; i<8; i++)
	fscanf(fid,"%f",&shape.az_angs[i+iaz*8+iorb*8*shape.nazs+ib*8*shape.nazs*shape.norbs]);

  for (iorb=0; iorb<shape.norbs; iorb++)
    for (iaz=0; iaz<shape.nazs; iaz++)
      for (islice=0; islice<8; islice++)
	for (i=0; i<16; i++)
	  fscanf(fid,"%f",&shape.corners[i+islice*16+iaz*8*16+iorb*8*16*shape.nazs+ib*8*16*shape.nazs*shape.norbs]);

  fclose(fid);
  return(0);
}



#define DTR 1.7453292519943295e-2
#define RTD 57.29577951308233

#define FLAT 3.3528131778969144e-3    /* = 1/298.257 FLATNESS, f, f=1-sqrt(1-e**2) */
#define AEARTH 6378.1363              /* SEMI-MAJOR AXIS OF EARTH, a, KM */


void get_shape(float time, float az, int ib, int islice, float cen_lat,
	       float cen_lon, int *nc, float *cor_lat, float *cor_lon, 
	       float orb_period)
{
  /*  given the orbit time, azimuth angle, beam, and slice number
      compute QuikScat slice shape using stored table

       DGL 7/12/199 (original fortran version) 
  */

  int i,j,ind1,ind2,ind3,ind4,io1,io2,ia1,ia2;
  float dela,delo,a;  
  static float table_orbit_period=6061.0;
  float t=table_orbit_period*(time/orb_period);

  float cx, sx, cd1, cd2, cd3, cd4, sd1, sd2, sd3, sd4, xx, yy;
  double r, rsq, temp, x[8], y[8];
  
  if (shape.norbs < 1) {  /* for the case when the file is not loaded */
    *nc=0;
    *cor_lat=0.0;
    *cor_lon=0.0;
    return;
  }  

  if (t < 0.0) t=t+table_orbit_period;
  if (t >= table_orbit_period) t=fmod(t,table_orbit_period);
  for (i=0; i<shape.norbs; i++){
    j=i+1;

    if (j+1 > shape.norbs) {
      j=j-shape.norbs;
      if (t >= shape.orbtimes[i] && t < shape.orbtimes[j]+table_orbit_period) {
	io1=i;
	io2=j;
	goto label_10;
      }
    } else {
      if (t >= shape.orbtimes[i] && t < shape.orbtimes[j]) {
	io1=i;
	io2=j;
	goto label_10;
      }
    }
  }
  
  printf("*** orbit time error in get_shape %f %f %f %d\n",time,t,orb_period,shape.norbs);
  exit(-1); /* error return */

label_10:

  delo=shape.orbtimes[io2]-shape.orbtimes[io1];
  if (io2 < io1) delo=delo+orb_period;
  delo=(t-shape.orbtimes[io1])/delo;
    
  a=fmod(az+720.0,360.0)*0.1;
  ia1=(int) a;
  ia2=ia1+1;    
  if (ia2 >= shape.nazs) ia2=ia2-shape.nazs;
  dela=a-ia1;
    
  /* bilinearly interpolate tables, first the azimuth angle */

  ind1=islice+ia1*8+io1*8*shape.nazs+ib*8*shape.nazs*shape.norbs;
  ind2=islice+ia1*8+io2*8*shape.nazs+ib*8*shape.nazs*shape.norbs;
  ind3=islice+ia2*8+io1*8*shape.nazs+ib*8*shape.nazs*shape.norbs;
  ind4=islice+ia2*8+io2*8*shape.nazs+ib*8*shape.nazs*shape.norbs;
  cd1=cos(DTR*shape.az_angs[ind1]);
  cd2=cos(DTR*shape.az_angs[ind2]);
  cd3=cos(DTR*shape.az_angs[ind3]);
  cd4=cos(DTR*shape.az_angs[ind4]);
  sd1=sin(DTR*shape.az_angs[ind1]);
  sd2=sin(DTR*shape.az_angs[ind2]);
  sd3=sin(DTR*shape.az_angs[ind3]);
  sd4=sin(DTR*shape.az_angs[ind4]);

  cx=rinterp(dela,delo,cd1,cd2,cd3,cd4);
  sx=rinterp(dela,delo,sd1,sd2,sd3,sd4);

  /* fix azimuth angle interpolation */
  cd1=sqrt(cx*cx+sx*sx);
  cx=cx/cd1;
  sx=sx/cd1;
  
  /* for each slice corner, interpolate and rotate */

  for (i=0; i<8; i++) {
    ind1=i+islice*16+ia1*8*16+io1*8*16*shape.nazs+ib*8*16*shape.nazs*shape.norbs;
    ind2=i+islice*16+ia1*8*16+io2*8*16*shape.nazs+ib*8*16*shape.nazs*shape.norbs;
    ind3=i+islice*16+ia2*8*16+io1*8*16*shape.nazs+ib*8*16*shape.nazs*shape.norbs;
    ind4=i+islice*16+ia2*8*16+io2*8*16*shape.nazs+ib*8*16*shape.nazs*shape.norbs;
    xx=rinterp(dela,delo,shape.corners[ind1], shape.corners[ind2], 
	       shape.corners[ind3], shape.corners[ind4]);
    yy=rinterp(dela,delo,shape.corners[ind1+8], shape.corners[ind2+8], 
	       shape.corners[ind3+8], shape.corners[ind4+8]);
    x[i] =  cx*xx + sx*yy;
    y[i] = -sx*xx + cx*yy;
  }

  /* convert x,y corner to lat,lon */
  xx=sin(DTR*cen_lat);
  r=(1.0-FLAT*xx*xx)*AEARTH;
  rsq=r*cos(DTR*cen_lat);
  
  for (i=0; i<8; i++) {
    temp=x[i]/rsq;
    if (temp > 1.0) temp = 1.0;
    if (temp < -1.0) temp = -1.0;
    temp=RTD*asin(temp);
    cor_lat[i]=RTD*asin((y[i]-(1.0-cos(DTR*temp))*sin(DTR*cen_lat)*rsq)/r)+cen_lat;
    cor_lon[i]=temp+cen_lon;
  }
  *nc=8;

  return;
}





